// ---------------------------------------------------------------------------
// Simple RegExp
//
//                                                 copyright (c) kanegon 2009
// ---------------------------------------------------------------------------

#include <tchar.h>
#include <windows.h>
#include <shlwapi.h>
#include <comdef.h>
#include "SimpleRegExp.h"

#pragma comment(lib, "shlwapi.lib")


//////////////////////////////////////////////////////////////////////////////

void SimpleRegExp::InitRegExp(LPCTSTR pszRegExp, LPCTSTR pszOption)
{
	Clear();
	if (FAILED(m_pRegExp.CreateInstance(__uuidof(RegExp)))) {
		throw regexp_exception("CreateInstance(RegExp) failed.");
	}
	m_pRegExp->PutPattern(pszRegExp);
	if (pszOption != NULL) {
		if (StrChr(pszOption, _T('g')) != NULL) {
			m_pRegExp->PutGlobal(VARIANT_TRUE);
		}
		if (StrChr(pszOption, _T('i')) != NULL) {
			m_pRegExp->PutIgnoreCase(VARIANT_TRUE);
		}
		if (StrChr(pszOption, _T('m')) != NULL) {
			m_pRegExp->PutMultiline(VARIANT_TRUE);
		}
	}
}

BOOL SimpleRegExp::Test(LPCTSTR pszText, LPCTSTR pszRegExp, LPCTSTR pszOption/*=NULL*/)
{
	InitRegExp(pszRegExp, pszOption);
	VARIANT_BOOL result = m_pRegExp->Test(pszText);
	return (result != VARIANT_FALSE);
}

BOOL SimpleRegExp::Match(LPCTSTR pszText, LPCTSTR pszRegExp, LPCTSTR pszOption/*=NULL*/)
{
	BOOL result = FALSE;
	InitRegExp(pszRegExp, pszOption);
	m_pMatches = m_pRegExp->Execute(pszText);
	if (m_pMatches->GetCount() > 0) {
		// select first match item by default.
		result = TRUE;
		m_pMatch = m_pMatches->GetItem(0);
		m_pSubMatches = m_pMatch->GetSubMatches();
	}
	return result;
}

xstr_t SimpleRegExp::Replace(LPCTSTR pszText, LPCTSTR pszReplaceText, LPCTSTR pszRegExp, LPCTSTR pszOption/*=NULL*/)
{
	InitRegExp(pszRegExp, pszOption);
	_variant_t replace = (_variant_t)(_bstr_t)pszReplaceText;
	return (LPCTSTR)(m_pRegExp->Replace(pszText, replace));
}

int SimpleRegExp::GetMatchCount()
{
	if (m_pMatches == NULL) {
		return 0;
	}
	return (int)(m_pMatches->GetCount());
}

BOOL SimpleRegExp::SelectMatch(int index=0)
{
	if (m_pMatches == NULL) {
		return FALSE;
	}
	if (m_pMatch != NULL) {
		m_pMatch = NULL;
		m_pSubMatches = NULL;
	}
	if (index < 0 || index >= m_pMatches->GetCount()) {
		return FALSE;
	}
	m_pMatch = m_pMatches->GetItem(index);
	m_pSubMatches = m_pMatch->GetSubMatches();
	return TRUE;
}

// index=0            text of total match
// index=1, 2, ...    text of match group ($1, $2, ...)
xstr_t SimpleRegExp::GetMatchText(int index/*=0*/)
{
	if (m_pMatch == NULL) {
		return _T("");
	}
	if (index == 0) {
		return (LPCTSTR)(m_pMatch->GetValue());
	}
	if (m_pSubMatches == NULL) {
		return _T("");
	}
	if (index < 0 || index > m_pSubMatches->GetCount()) {
		return _T("");
	}
	return (LPCTSTR)(_bstr_t)(m_pSubMatches->GetItem(index - 1));
}

