//----------------------------------------------------------------------------
// simple string
//
//                                            copyright (c) kanegon 2003-2009
//----------------------------------------------------------------------------

#pragma warning (disable: 4995 4996)

#include <tchar.h>
#include <windows.h>
#include <shlwapi.h>
#include <new>
#include <string.h>
#include "xstr.h"

#pragma comment(lib, "shlwapi.lib")


//////////////////////////////////////////////////////////////////////////////
// helper function

char* wstr2mb(const wchar_t* pwsz)
{
	if (pwsz == NULL) {
		return NULL;
	}
	int len = WideCharToMultiByte(CP_ACP, 0, pwsz, -1, NULL, 0, NULL, NULL);
	char* psz = new char[len];
	WideCharToMultiByte(CP_ACP, 0, pwsz, -1, psz, len, NULL, NULL);
	return psz;
}

wchar_t* mb2wstr(const char* psz)
{
	if (psz == NULL) {
		return NULL;
	}
	int len = MultiByteToWideChar(CP_ACP, 0, psz, -1, NULL, 0);
	wchar_t* wstr = new wchar_t[len];
	MultiByteToWideChar(CP_ACP, 0, psz, -1, wstr, len);
    return wstr;
}

BSTR mb2bstr(const char* psz)
{
	if (psz == NULL) {
		return NULL;
	}
	int len = MultiByteToWideChar(CP_ACP, 0, psz, -1, NULL, 0);
	BSTR bstr = SysAllocStringLen(NULL, len - 1);
	if (bstr == NULL) {
		throw std::bad_alloc();
	}
	MultiByteToWideChar(CP_ACP, 0, psz, -1, bstr, len);
    return bstr;
}


//////////////////////////////////////////////////////////////////////////////

char* wstr2mb_no_allocate(char* buff, size_t buff_len, const wchar_t* pwsz)
{
	buff[0] = '\0';
	if (pwsz == NULL) {
		return buff;
	}
	int len = WideCharToMultiByte(CP_ACP, 0, pwsz, -1, NULL, 0, NULL, NULL);
	if (len > (int)buff_len) {
		return buff;
	}
	WideCharToMultiByte(CP_ACP, 0, pwsz, -1, buff, len, NULL, NULL);
	return buff;
}

wchar_t* mb2wstr_no_allocate(wchar_t* buff, size_t buff_len, const char* psz)
{
	buff[0] = L'\0';
	if (psz == NULL) {
		return buff;
	}
	int len = MultiByteToWideChar(CP_ACP, 0, psz, -1, NULL, 0);
	if (len > (int)buff_len) {
		return buff;
	}
	MultiByteToWideChar(CP_ACP, 0, psz, -1, buff, len);
    return buff;
}


//////////////////////////////////////////////////////////////////////////////

const char* StrRStrA(const char* pszSource, const char* pszLast, const char* pszSrch)
{
	if (pszSource == NULL || pszSrch == NULL) {
		return NULL;
	}
	int nSrchTextLen = lstrlenA(pszSrch);
	const char* pszFindStr = NULL;
	const char* pszLastTmp;
	if (pszLast != NULL) {
		pszLastTmp = pszLast;
	} else {
		pszLastTmp = pszSource + lstrlenA(pszSource);
	}
	for (;;) {
		pszFindStr = StrRStrIA(pszSource, pszLastTmp, pszSrch);
		if (pszFindStr == NULL) {
			break;
		}
		if (memcmp(pszFindStr, pszSrch, nSrchTextLen) == 0) {
			break;
		}
		pszLastTmp = pszFindStr;
	}
	return pszFindStr;
}

const wchar_t* StrRStrW(const wchar_t* pszSource, const wchar_t* pszLast, const wchar_t* pszSrch)
{
	if (pszSource == NULL || pszSrch == NULL) {
		return NULL;
	}
	int nSrchTextLen = lstrlenW(pszSrch) * sizeof(WCHAR);
	const wchar_t* pszFindStr = NULL;
	const wchar_t* pszLastTmp;
	if (pszLast != NULL) {
		pszLastTmp = pszLast;
	} else {
		pszLastTmp = pszSource + lstrlenW(pszSource);
	}
	for (;;) {
		pszFindStr = StrRStrIW(pszSource, pszLastTmp, pszSrch);
		if (pszFindStr == NULL) {
			break;
		}
		if (memcmp(pszFindStr, pszSrch, nSrchTextLen) == 0) {
			break;
		}
		pszLastTmp = pszFindStr;
	}
	return pszFindStr;
}


#ifndef USE_STD_STRING

//////////////////////////////////////////////////////////////////////////////
// xstr

xstr::xstr()
{
	m_pszBuffer = NULL;
	m_bAllowNull = false;
}

xstr::xstr(const char* psz, bool allowNull/*=false*/)
{
	assign_raw(psz);
	m_bAllowNull = allowNull;
}

xstr::xstr(const wchar_t* pwsz, bool allowNull/*=false*/)
{
	char* psz = NULL;
	if (pwsz != NULL) {
		psz = wstr2mb(pwsz);
	}
	assign_raw(psz);
	if (psz != NULL) {
		delete[] psz;
	}
	m_bAllowNull = allowNull;
}

xstr::xstr(const xstr& str)
{
	if (str.m_pszBuffer == NULL) {
		m_pszBuffer = NULL;
	} else {
		m_pszBuffer = new char[strlen(str.m_pszBuffer) + 1];
		strcpy(m_pszBuffer, str.m_pszBuffer);
	}
	m_bAllowNull = str.m_bAllowNull;
}

xstr::~xstr()
{
	clear();
}

void xstr::assign_raw(const char* psz)
{
	if (psz == NULL) {
		m_pszBuffer = NULL;
	} else {
		m_pszBuffer = new char[strlen(psz) + 1];
		strcpy(m_pszBuffer, psz);
	}
}

void xstr::clear()
{
	if (m_pszBuffer != NULL) {
		delete[] m_pszBuffer;
	}
	m_pszBuffer = NULL;
}

xstr& xstr::assign(const char* psz)
{
	if (psz == m_pszBuffer) {
		return *this;
	}
	clear();
	assign_raw(psz);
	return *this;
}

xstr& xstr::concat(const char* psz)
{
	if (psz == NULL) {
		return *this;
	}
	if (m_pszBuffer == NULL) {
		assign_raw(psz);
	}
	else {
		int len = strlen(m_pszBuffer) + strlen(psz);
		char* pszNewBuffer = new char[len + 1];
		strcpy(pszNewBuffer, m_pszBuffer);
		strcat(pszNewBuffer, psz);
		delete[] m_pszBuffer;
		m_pszBuffer = pszNewBuffer;
	}
	return *this;
}

xstr& xstr::operator+=(const char* psz)
{
	return concat(psz);
}

xstr& xstr::operator=(const char* psz)
{
	// m_bAllowNull attrinute not changed.
	return assign(psz);
}

xstr& xstr::operator=(const xstr& str)
{
	// m_bAllowNull attrinute not changed.
	return assign(str);
}

xstr::operator const char*() const
{
	return c_str();
}

const char* xstr::c_str() const
{
	if (m_pszBuffer == NULL && ! m_bAllowNull) {
		return "";
	}
	return m_pszBuffer;
}

int xstr::length() const
{
	return (m_pszBuffer != NULL) ? strlen(m_pszBuffer) : 0;
}

bool xstr::empty() const
{
	return (m_pszBuffer == NULL || m_pszBuffer[0] == '\0');
}

bool xstr::null() const
{
	return (m_bAllowNull && (m_pszBuffer == NULL));
}

char* xstr::getBuffer(int size/*=0*/)
{
	if (size < 0) {
		size = 0;
	}
	int len = length();
	if (m_pszBuffer == NULL || size > len) {
		char* pszNewBuffer = new char[size + 1];
		ZeroMemory(pszNewBuffer, size + 1);
		if (m_pszBuffer != NULL) {
			strcpy(pszNewBuffer, m_pszBuffer);
			delete[] m_pszBuffer;
		}
		m_pszBuffer = pszNewBuffer;
	}
	return m_pszBuffer;
}

int xstr::compare(const char* psz) const
{
	if (empty()) {
		if (psz == NULL || psz[0] == '\0') {
			return 0;
		}
		return -1;
	}
	if (psz == NULL || psz[0] == '\0') {
		return 1;
	}
	return StrCmpA(m_pszBuffer, psz);
}

int xstr::find(char ch, int start/*=0*/) const
{
	if (m_pszBuffer == NULL) {
		return -1;
	}
	if (start < 0) {
		start = 0;
	}
	if (start >= (int)strlen(m_pszBuffer)) {
		return -1;
	}
	const char* pszFind = StrChrA(m_pszBuffer + start, ch);
	if (pszFind == NULL) {
		return -1;
	}
	return (pszFind - m_pszBuffer);
}

int xstr::find(char* psz, int start/*=0*/) const
{
	if (m_pszBuffer == NULL || psz == NULL) {
		return -1;
	}
	if (start < 0) {
		start = 0;
	}
	if (start >= (int)strlen(m_pszBuffer)) {
		return -1;
	}
	const char* pszFind = StrStrA(m_pszBuffer + start, psz);
	if (pszFind == NULL) {
		return -1;
	}
	return (pszFind - m_pszBuffer);
}

int xstr::rfind(char ch, int start/*=-1*/) const
{
	if (m_pszBuffer == NULL) {
		return -1;
	}
	int len = strlen(m_pszBuffer);
	if (start < 0 || start > len) {
		start = len;
	}
	const char* pszFind = StrRChrA(m_pszBuffer, m_pszBuffer + start, ch);
	if (pszFind == NULL) {
		return -1;
	}
	return (pszFind - m_pszBuffer);
}

int xstr::rfind(char* psz, int start/*=-1*/) const
{
	if (m_pszBuffer == NULL) {
		return -1;
	}
	int len = strlen(m_pszBuffer);
	if (start < 0 || start > len) {
		start = len;
	}
	const char* pszFind = StrRStrA(m_pszBuffer, m_pszBuffer + start, psz);
	if (pszFind == NULL) {
		return -1;
	}
	return (pszFind - m_pszBuffer);
}

xstr xstr::left(int len) const
{
	return left(m_pszBuffer, len);
}

xstr xstr::right(int len) const
{
	return right(m_pszBuffer, len);
}

xstr xstr::substr(int start, int len/*=-1*/) const
{
	return substr(m_pszBuffer, start, len);
}

xstr& xstr::trim()
{
	*this = trim(m_pszBuffer);
	return *this;
}

xstr xstr::left(const char* psz, int len)
{
	if (psz == NULL || len < 0)
		return "";
	if (len > (int)strlen(psz))
		len = strlen(psz);
	xstr str;
	char* pszBuffer = str.getBuffer(len);
	memcpy(pszBuffer, psz, len);
	pszBuffer[len] = '\0';
	return str;
}

xstr xstr::right(const char* psz, int len)
{
	if (psz == NULL || len < 0) {
		return "";
	}
	if (len > (int)strlen(psz)) {
		len = strlen(psz);
	}
	xstr str;
	char* pszBuffer = str.getBuffer(len);
	memcpy(pszBuffer, &psz[strlen(psz) - len], len);
	pszBuffer[len] = '\0';
	return str;
}

xstr xstr::substr(const char* psz, int start, int len/*=-1*/)
{
	if (psz == NULL) {
		return "";
	}
	int slen = strlen(psz);
	int end;
	if (len >= 0) {
		end = start + len;
	}
	else {
		end = slen;
	}
	if (start < 0) {
		start = 0;
	}
	if (start > slen) {
		start = slen;
	}
	if (end < start) {
		end = start;
	}
	if (end > slen) {
		end = slen;
	}
	len = end - start;
	xstr str;
	char* pszBuffer = str.getBuffer(len);
	memcpy(pszBuffer, &psz[start], len);
	pszBuffer[len] = '\0';
	return str;
}

xstr xstr::trim(const char* psz)
{
	if (psz == NULL) {
		return "";
	}
	int slen = strlen(psz);
	int start = 0;
	int end = 0;
	int index = 0;
	for (; index < slen; index++) {
		if (! isspace((BYTE)psz[index])) {
			break;
		}
		start = index + 1;
	}
	end = start;
	for (; index < slen; index++) {
		if (! isspace((BYTE)psz[index])) {
			end = index + 1;
		}
	}
	int len = end - start;
	return xstr::substr(psz, start, len);
}

xstr operator+(const xstr& str, const char* psz)
{
	xstr newstr(str);
	newstr.concat(psz);
	return newstr;
}


//////////////////////////////////////////////////////////////////////////////
// xwstr

xwstr::xwstr()
{
	m_pszBuffer = NULL;
	m_bAllowNull = false;
}

xwstr::xwstr(const wchar_t* psz, bool allowNull/*=false*/)
{
	assign_raw(psz);
	m_bAllowNull = allowNull;
}

xwstr::xwstr(const char* psz, bool allowNull/*=false*/)
{
	wchar_t* pwsz = NULL;
	if (psz != NULL) {
		pwsz = mb2wstr(psz);
	}
	assign_raw(pwsz);
	if (pwsz != NULL) {
		delete[] pwsz;
	}
	m_bAllowNull = allowNull;
}

xwstr::xwstr(const xwstr& str)
{
	if (str.m_pszBuffer == NULL) {
		m_pszBuffer = NULL;
	} else {
		m_pszBuffer = new wchar_t[wcslen(str.m_pszBuffer) + 1];
		wcscpy(m_pszBuffer, str.m_pszBuffer);
	}
	m_bAllowNull = str.m_bAllowNull;
}

xwstr::~xwstr()
{
	clear();
}

void xwstr::assign_raw(const wchar_t* psz)
{
	if (psz == NULL) {
		m_pszBuffer = NULL;
	} else {
		m_pszBuffer = new wchar_t[wcslen(psz) + 1];
		wcscpy(m_pszBuffer, psz);
	}
}

void xwstr::clear()
{
	if (m_pszBuffer != NULL) {
		delete[] m_pszBuffer;
	}
	m_pszBuffer = NULL;
}

xwstr& xwstr::assign(const wchar_t* psz)
{
	if (psz == m_pszBuffer) {
		return *this;
	}
	clear();
	assign_raw(psz);
	return *this;
}

xwstr& xwstr::concat(const wchar_t* psz)
{
	if (psz == NULL) {
		return *this;
	}
	if (m_pszBuffer == NULL) {
		assign_raw(psz);
	}
	else {
		int len = wcslen(m_pszBuffer) + wcslen(psz);
		wchar_t* pszNewBuffer = new wchar_t[len + 1];
		wcscpy(pszNewBuffer, m_pszBuffer);
		wcscat(pszNewBuffer, psz);
		delete[] m_pszBuffer;
		m_pszBuffer = pszNewBuffer;
	}
	return *this;
}

xwstr& xwstr::operator+=(const wchar_t* psz)
{
	return concat(psz);
}

xwstr& xwstr::operator=(const wchar_t* psz)
{
	return assign(psz);
}

xwstr& xwstr::operator=(const xwstr& str)
{
	return assign(str);
}

xwstr::operator const wchar_t*() const
{
	return c_str();
}

const wchar_t* xwstr::c_str() const
{
	if (m_pszBuffer == NULL && ! m_bAllowNull) {
		return L"";
	}
	return m_pszBuffer;
}

int xwstr::length() const
{
	return (m_pszBuffer != NULL) ? wcslen(m_pszBuffer) : 0;
}

bool xwstr::empty() const
{
	return (m_pszBuffer == NULL || m_pszBuffer[0] == L'\0');
}

bool xwstr::null() const
{
	return (m_bAllowNull && (m_pszBuffer == NULL));
}

wchar_t* xwstr::getBuffer(int size/*=0*/)
{
	if (size < 0) {
		size = 0;
	}
	int len = length();
	if (m_pszBuffer == NULL || size > len) {
		wchar_t* pszNewBuffer = new wchar_t[size + 1];
		ZeroMemory(pszNewBuffer, (size + 1) * sizeof(wchar_t));
		if (m_pszBuffer != NULL) {
			wcscpy(pszNewBuffer, m_pszBuffer);
			delete[] m_pszBuffer;
		}
		m_pszBuffer = pszNewBuffer;
	}
	return m_pszBuffer;
}

int xwstr::compare(const wchar_t* psz) const
{
	if (empty()) {
		if (psz == NULL || psz[0] == L'\0') {
			return 0;
		}
		return -1;
	}
	if (psz == NULL || psz[0] == L'\0') {
		return 1;
	}
	return StrCmpW(m_pszBuffer, psz);
}

int xwstr::find(wchar_t ch, int start/*=0*/) const
{
	if (m_pszBuffer == NULL) {
		return -1;
	}
	if (start < 0) {
		start = 0;
	}
	if (start >= (int)wcslen(m_pszBuffer)) {
		return -1;
	}
	const wchar_t* pszFind = StrChrW(m_pszBuffer + start, ch);
	if (pszFind == NULL) {
		return -1;
	}
	return (pszFind - m_pszBuffer);
}

int xwstr::find(wchar_t* psz, int start/*=0*/) const
{
	if (m_pszBuffer == NULL || psz == NULL) {
		return -1;
	}
	if (start < 0) {
		start = 0;
	}
	if (start >= (int)wcslen(m_pszBuffer)) {
		return -1;
	}
	const wchar_t* pszFind = StrStrW(m_pszBuffer + start, psz);
	if (pszFind == NULL) {
		return -1;
	}
	return (pszFind - m_pszBuffer);
}

int xwstr::rfind(wchar_t ch, int start/*=-1*/) const
{
	if (m_pszBuffer == NULL) {
		return -1;
	}
	int len = wcslen(m_pszBuffer);
	if (start < 0 || start > len) {
		start = len;
	}
	const wchar_t* pszFind = StrRChrW(m_pszBuffer, m_pszBuffer + start, ch);
	if (pszFind == NULL) {
		return -1;
	}
	return (pszFind - m_pszBuffer);
}

int xwstr::rfind(wchar_t* psz, int start/*=-1*/) const
{
	if (m_pszBuffer == NULL) {
		return -1;
	}
	int len = wcslen(m_pszBuffer);
	if (start < 0 || start > len) {
		start = len;
	}
	const wchar_t* pszFind = StrRStrW(m_pszBuffer, m_pszBuffer + start, psz);
	if (pszFind == NULL) {
		return -1;
	}
	return (pszFind - m_pszBuffer);
}

xwstr xwstr::left(int len) const
{
	return left(m_pszBuffer, len);
}

xwstr xwstr::right(int len) const
{
	return right(m_pszBuffer, len);
}

xwstr xwstr::substr(int start, int len/*=-1*/)
{
	return substr(m_pszBuffer, start, len);
}

xwstr& xwstr::trim()
{
	*this = trim(m_pszBuffer);
	return *this;
}

xwstr xwstr::left(const wchar_t* psz, int len)
{
	if (psz == NULL || len < 0)
		return L"";
	if (len > (int)wcslen(psz))
		len = (int)wcslen(psz);
	xwstr str;
	wchar_t* pszBuffer = str.getBuffer(len);
	memcpy(pszBuffer, psz, len * sizeof(wchar_t));
	pszBuffer[len] = L'\0';
	return str;
}

xwstr xwstr::right(const wchar_t* psz, int len)
{
	if (psz == NULL || len < 0) {
		return "";
	}
	if (len > (int)wcslen(psz)) {
		len = wcslen(psz);
	}
	xwstr str;
	wchar_t* pszBuffer = str.getBuffer(len);
	memcpy(pszBuffer, &psz[wcslen(psz) - len], len * sizeof(wchar_t));
	pszBuffer[len] = L'\0';
	return str;
}

xwstr xwstr::substr(const wchar_t* psz, int start, int len/*=-1*/)
{
	if (psz == NULL) {
		return L"";
	}
	int slen = wcslen(psz);
	int end;
	if (len >= 0) {
		end = start + len;
	}
	else {
		end = slen;
	}
	if (start < 0) {
		start = 0;
	}
	if (start > slen) {
		start = slen;
	}
	if (end < start) {
		end = start;
	}
	if (end > slen) {
		end = slen;
	}
	len = end - start;
	xwstr str;
	wchar_t* pszBuffer = str.getBuffer(len);
	memcpy(pszBuffer, &psz[start], len * sizeof(wchar_t));
	pszBuffer[len] = L'\0';
	return str;
}

xwstr xwstr::trim(const wchar_t* psz)
{
	if (psz == NULL) {
		return L"";
	}
	int slen = wcslen(psz);
	int start = 0;
	int end = 0;
	int index = 0;
	for (; index < slen; index++) {
		if (! iswspace((WORD)psz[index])) {
			break;
		}
		start = index + 1;
	}
	end = start;
	for (; index < slen; index++) {
		if (! iswspace((WORD)psz[index])) {
			end = index + 1;
		}
	}
	int len = end - start;
	return xwstr::substr(psz, start, len);
}

xwstr operator+(const xwstr& str, const wchar_t* psz)
{
	xwstr newstr(str);
	newstr.concat(psz);
	return newstr;
}


//////////////////////////////////////////////////////////////////////////////
// xbstr

xbstr::xbstr()
{
	m_pszBuffer = NULL;
	m_pszConvertText = NULL;
}

xbstr::xbstr(const char* psz)
{
	m_pszBuffer = NULL;
	m_pszConvertText = NULL;
	assign_raw(psz);
}

xbstr::xbstr(wchar_t* psz, bool attach/*=false*/)
{
	m_pszBuffer = NULL;
	m_pszConvertText = NULL;
	if (attach && psz != NULL) {
		// psz must be BSTR
		m_pszBuffer = psz;
	} else {
		assign_raw(psz);
	}
}

xbstr::xbstr(const wchar_t* psz)
{
	m_pszBuffer = NULL;
	m_pszConvertText = NULL;
	assign_raw(psz);
}

xbstr::xbstr(const xbstr& str)
{
	m_pszBuffer = NULL;
	m_pszConvertText = NULL;
	assign_raw(str.m_pszBuffer);
}

xbstr::~xbstr()
{
	clear();
}

void xbstr::assign_raw(const char* psz)
{
	if (m_pszBuffer != NULL) {
		clear();
	}
	if (psz != NULL) {
		m_pszBuffer = mb2bstr(psz);
	}
}

void xbstr::assign_raw(const wchar_t* psz)
{
	if (m_pszBuffer != NULL) {
		clear();
	}
	if (psz != NULL) {
		m_pszBuffer = SysAllocString(psz);
		if (m_pszBuffer == NULL) {
			throw std::bad_alloc();
		}
	}
}

void xbstr::clear()
{
	if (m_pszBuffer != NULL) {
		SysFreeString(m_pszBuffer);
		m_pszBuffer = NULL;
	}
	if (m_pszConvertText != NULL) {
		delete[] m_pszConvertText;
		m_pszConvertText = NULL;
	}
}

xbstr& xbstr::assign(const char* psz)
{
	if (psz == m_pszConvertText) {
		return *this;
	}
	assign_raw(psz);
	return *this;
}

xbstr& xbstr::assign(const wchar_t* psz)
{
	if (psz == m_pszBuffer) {
		return *this;
	}
	assign_raw(psz);
	return *this;
}

xbstr& xbstr::operator=(const char* psz)
{
	return assign(psz);
}

xbstr& xbstr::operator=(const wchar_t* psz)
{
	return assign(psz);
}

xbstr& xbstr::operator=(const xbstr& str)
{
	return assign(str.m_pszBuffer);
}

xbstr::operator const char*()
{
	if (m_pszConvertText == NULL) {
		m_pszConvertText = wstr2mb(m_pszBuffer);
	}
	return m_pszConvertText;
}

xbstr::operator const wchar_t*() const
{
	return m_pszBuffer;
}

int xbstr::length() const
{
	if (m_pszBuffer == NULL) {
		return 0;
	}
	return SysStringLen(m_pszBuffer);
}

bool xbstr::empty() const
{
	return (length() == 0);
}

bool xbstr::null() const
{
	return (m_pszBuffer == NULL);
}

wchar_t* xbstr::getBuffer(int size/*=0*/)
{
	int len = 0;
	if (m_pszBuffer != NULL) {
		len = SysStringLen(m_pszBuffer);
	}
	int newlen = (len >= size) ? len : size;
	BSTR bstr = SysAllocStringLen(NULL, newlen);
	if (bstr == NULL) {
		throw std::bad_alloc();
	}
	ZeroMemory(bstr, newlen * sizeof(wchar_t) + 1);
	memcpy(bstr, m_pszBuffer, len * sizeof(wchar_t));
	clear();
	m_pszBuffer = bstr;
	return m_pszBuffer;
}

#endif // USE_STD_STRING
