//----------------------------------------------------------------------------
// simple string
//
//                                            copyright (c) kanegon 2003-2009
//----------------------------------------------------------------------------

#ifndef _XSTR_H_
#define _XSTR_H_

#include <tchar.h>
#include <windows.h>
#include <comdef.h>
#include <string>


//////////////////////////////////////////////////////////////////////////////
// helper function

char* wstr2mb(const wchar_t* pwsz);
wchar_t* mb2wstr(const char* psz);
BSTR mb2bstr(const char* psz);

char* wstr2mb_no_allocate(char* buff, size_t buff_len, const wchar_t* pwsz);
wchar_t* mb2wstr_no_allocate(wchar_t* buff, size_t buff_len, const char* psz);


//////////////////////////////////////////////////////////////////////////////

LPCSTR StrRStrA(LPCSTR lpSource, LPCSTR lpLast, LPCSTR lpSrch);
LPCWSTR StrRStrW(LPCWSTR lpSource, LPCWSTR lpLast, LPCWSTR lpSrch);

#ifdef UNICODE
#define StrRStr                StrRStrW
#else
#define StrRStr                StrRStrA
#endif


//////////////////////////////////////////////////////////////////////////////

#ifdef USE_STD_STRING

typedef std::basic_string<char>    xstr;
typedef std::basic_string<wchar_t> xwstr;

#ifdef _UNICODE
typedef xwstr xstr_t;
#else
typedef xstr  xstr_t;
#endif

#else // USE_STD_STRING

//////////////////////////////////////////////////////////////////////////////
// simple string (char*)

class xstr
{
protected:
	char* m_pszBuffer;
	bool m_bAllowNull;

public:
	xstr();
	xstr(const char* psz, bool allowNull = false);
	xstr(const wchar_t* pwsz, bool allowNull = false);
	xstr(const xstr& str);
	virtual ~xstr();

protected:
	void assign_raw(const char* psz);

public:
	void clear();
	xstr& assign(const char* psz);
	xstr& concat(const char* psz);
	xstr& operator+=(const char* psz);
	xstr& operator=(const char* psz);
	xstr& operator=(const xstr& str);
	operator const char*() const;
	const char* c_str() const;

	int length() const;
	bool empty() const;
	bool null() const;

	char* getBuffer(int size=0);

	int compare(const char* psz) const;
	int find(char ch, int start=0) const;
	int find(char* psz, int start=0) const;
	int rfind(char ch, int start=0) const;
	int rfind(char* psz, int start=0) const;

	xstr left(int len) const;
	xstr right(int len) const;
	xstr substr(int start, int len=-1) const;
	xstr& trim();

	static xstr left(const char* psz, int len);
	static xstr right(const char* psz, int len);
	static xstr substr(const char* psz, int start, int len=-1);
	static xstr trim(const char* psz);

	friend xstr operator+(const xstr& str, const char* psz);
};


//////////////////////////////////////////////////////////////////////////////
// simple string (wchar_t*)

class xwstr
{
protected:
	wchar_t* m_pszBuffer;
	bool m_bAllowNull;

public:
	xwstr();
	xwstr(const wchar_t* psz, bool allowNull = FALSE);
	xwstr(const char* psz, bool allowNull = FALSE);
	xwstr(const xwstr& str);
	virtual ~xwstr();

protected:
	void assign_raw(const wchar_t* psz);

public:
	void clear();
	xwstr& assign(const wchar_t* psz);
	xwstr& concat(const wchar_t* psz);
	xwstr& operator+=(const wchar_t* psz);
	xwstr& operator=(const wchar_t* psz);
	xwstr& operator=(const xwstr& str);
	operator const wchar_t*() const;
	const wchar_t* c_str() const;

	int length() const;
	bool empty() const;
	bool null() const;

	wchar_t* getBuffer(int size=0);

	int compare(const wchar_t* psz) const;
	int find(wchar_t ch, int start=0) const;
	int find(wchar_t* psz, int start=0) const;
	int rfind(wchar_t ch, int start=0) const;
	int rfind(wchar_t* psz, int start=0) const;

	xwstr left(int len) const;
	xwstr right(int len) const;
	xwstr substr(int start, int len=-1);
	xwstr& trim();

	static xwstr left(const wchar_t* psz, int len);
	static xwstr right(const wchar_t* psz, int len);
	static xwstr substr(const wchar_t* psz, int start, int len=-1);
	static xwstr trim(const wchar_t* psz);

	friend xwstr operator+(const xwstr& str, const wchar_t* psz);
};


//////////////////////////////////////////////////////////////////////////////

#if defined(UNICODE)
typedef xwstr  xstr_t;
#else
typedef xstr   xstr_t;
#endif


//////////////////////////////////////////////////////////////////////////////
// simple string (BSTR)

class xbstr {
protected:
	BSTR m_pszBuffer;
	char* m_pszConvertText;

public:
	xbstr();
	xbstr(const char* psz);
	xbstr(wchar_t* psz, bool attach = false);
	xbstr(const wchar_t* psz);
	xbstr(const xbstr& str);
	virtual ~xbstr();

protected:
	void assign_raw(const char* psz);
	void assign_raw(const wchar_t* psz);

public:
	void clear();
	xbstr& assign(const char* psz);
	xbstr& assign(const wchar_t* psz);
	xbstr& operator=(const char* psz);
	xbstr& operator=(const wchar_t* psz);
	xbstr& operator=(const xbstr& str);
	operator const char*();
	operator const wchar_t*() const;

	int length() const;
	bool empty() const;
	bool null() const;

	wchar_t* getBuffer(int size=0);
};

#endif // USE_STD_STRING

#endif // _XSTR_H_
